#import "msado15.dll" rename("EOF", "EndOfFile")

#include <string>
#include <sstream>
#include <iostream>
#include <memory>

using namespace ADODB;
using namespace std;

const char *gConnectionString = "DSN=mysql_odbc_memory_corruption";
const char *gTableName = "large_varchars";
const int gVarCharSize = 7000;
const int gNumRowsToGenerate = 250;


static inline void TESTHR(HRESULT x) { if FAILED(x) _com_issue_error(x); };

static void GenerateVarCharString(char buff[], int buffSize, int row)
{
	int startPos = sprintf_s(buff, buffSize, "%d ", row);	// Add a number for easy tracking
	for (int i = startPos; i < buffSize - 1; i++)
		buff[i] = 'A' + ((i - startPos) % ('Z' - 'A' + 1));
	buff[buffSize - 1] = '\0';
}

static void DropTableIfExist(_ConnectionPtr connection, const std::string &tableName)
{
	connection->Execute(("DROP TABLE IF EXISTS " + tableName).c_str(), nullptr, -1);
}

static void CreateTestTable(_ConnectionPtr connection, const std::string &tableName, int varCharSize)
{
	// Generate sql string
	std::ostringstream sqlStringBuilder;
	sqlStringBuilder << "CREATE TABLE " << tableName << " (";
	sqlStringBuilder << "Col0 VARCHAR(" << varCharSize << ") CHARACTER SET ascii";
	sqlStringBuilder << ")";

	connection->Execute(sqlStringBuilder.str().c_str(), nullptr, -1);
}

static void PopulateTestTable(_ConnectionPtr connection, const std::string &tableName, int numRowsToGenerate, int varCharSize)
{
	_RecordsetPtr recordSet = nullptr;
	TESTHR(recordSet.CreateInstance(__uuidof(Recordset)));

	recordSet->Open(tableName.c_str(), _variant_t((IDispatch *)connection), adOpenKeyset, adLockOptimistic, -1);

	auto valBuff = std::make_unique<char[]>(varCharSize);
	for (int rowIdx = 0; rowIdx < numRowsToGenerate; rowIdx++) {
		GenerateVarCharString(valBuff.get(), varCharSize, rowIdx + 1);
		recordSet->AddNew("Col0", valBuff.get());
	}
}

static void GenerateTestTable(_ConnectionPtr connection, const std::string &tableName)
{
	DropTableIfExist(connection, tableName);
	CreateTestTable(connection, tableName, gVarCharSize);
	PopulateTestTable(connection, tableName, gNumRowsToGenerate, gVarCharSize);
}

static void TestRecordSetMove(_ConnectionPtr connection, const std::string &tableName)
{
	_RecordsetPtr recordSet;
	TESTHR(recordSet.CreateInstance(__uuidof(Recordset)));

	recordSet->Open(("SELECT Col0 FROM " + tableName).c_str(), _variant_t((IDispatch *)connection), adOpenKeyset, adLockOptimistic, -1);
	recordSet->MoveFirst();
	recordSet->Move(20);		// Causes write access violation in mysql odbc driver
}

void main()
{
	CoInitialize(0);

	try {
		_ConnectionPtr connection = nullptr;
		TESTHR(connection.CreateInstance(__uuidof(Connection)));
		connection->Open(gConnectionString, "", "", -1);

		GenerateTestTable(connection, gTableName);
		TestRecordSetMove(connection, gTableName);
	} catch (_com_error &e) {
		cout << "Error:";
		cout << e.ErrorMessage() << std::endl;
		cout << e.Description() << std::endl;
	}

	CoUninitialize();
}